unit UCpu;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils;

type
  TOpcode = (opNop, opHalt, opLoad, opLoadi, opJump, opJumpZero, opJumpNotZero,
    opJumpRel, opJumpRelZero, opJumpRelNotZero, opNeg, opClear, opLoadMem,
    opStoreMem, opExchg, opPush, opPop, opCall, opRet, opAdd, opAddi, opSub,
    opSubi, opInc, opDec, opIn, opOut, opShl, opShr, opDataPrefix8, opDataPrefix16,
    opDataPrefix32, opDataPrefix64, opDataSize, opAddrSize, opTest, opAnd, opOr,
    opXor, opLddr, opLdir, opMul, opDiv, opMod, opAddrPrefix8, opAddrPrefix16,
    opAddrPrefix32, opAddrPrefix64, opConvert);

  TAddressSigned = Int64;
  TAddress = QWord;
  PAddress = ^TAddress;
  TRegIndex = Byte;

  TBitWidth = (bwNone, bw8, bw16, bw32, bw64);
  TRegister = record
    case TBitWidth of
      bw8: (B: Byte);
      bw16: (W: Word);
      bw32: (D: DWord);
      bw64: (Q: QWord);
  end;

  TInstructionEvent = procedure of object;
  TInputEvent = function (Port: TAddress): TRegister of object;
  TOutputEvent = procedure (Port: TAddress; Value: TRegister) of object;

  TCpuThread = class;

  { TCpu }

  TCpu = class
  private
    FAddrSizeBase: TBitWidth;
    FDataSizeBase: TBitWidth;
    FOnInput: TInputEvent;
    FOnOutput: TOutputEvent;
    FRunning: Boolean;
    FTicks: Integer;
    Instructions: array[TOpcode] of TInstructionEvent;
    Prefix: Boolean;
    Z: Boolean;
    Thread: TCpuThread;
    procedure InstConvert;
    procedure InstNop;
    procedure InstHalt;
    procedure InstLoad;
    procedure InstLoadImmediate;
    procedure InstJump;
    procedure InstJumpZero;
    procedure InstJumpNotZero;
    procedure InstJumpRel;
    procedure InstJumpRelZero;
    procedure InstJumpRelNotZero;
    procedure InstTest;
    procedure InstNeg;
    procedure InstClear;
    procedure InstLoadMem;
    procedure InstStoreMem;
    procedure InstExchg;
    procedure InstPush;
    procedure InstPop;
    procedure InstCall;
    procedure InstRet;
    procedure InstAdd;
    procedure InstAddi;
    procedure InstSub;
    procedure InstSubi;
    procedure InstMul;
    procedure InstDiv;
    procedure InstMod;
    procedure InstInc;
    procedure InstDec;
    procedure InstIn;
    procedure InstOut;
    procedure InstShr;
    procedure InstShl;
    procedure InstAnd;
    procedure InstOr;
    procedure InstXor;
    procedure InstLddr;
    procedure InstLdir;
    procedure InstDataPrefix8;
    procedure InstDataPrefix16;
    procedure InstDataPrefix32;
    procedure InstDataPrefix64;
    procedure InstDataSize;
    procedure InstAddrPrefix8;
    procedure InstAddrPrefix16;
    procedure InstAddrPrefix32;
    procedure InstAddrPrefix64;
    procedure InstAddrSize;
    procedure InitInstructions;
    procedure SetAddrSizeBase(AValue: TBitWidth);
    procedure SetDataSizeBase(AValue: TBitWidth);
  public
    Memory: Pointer;
    Registers: array of TRegister;
    Terminated: Boolean;
    IP: TAddress;
    SP: TAddress;
    DataSize: TBitWidth;
    AddrSize: TBitWidth;
    procedure Run;
    procedure Step; inline;
    procedure Start;
    procedure Stop;
    function Read8: Byte; inline;
    function Read16: Word; inline;
    function Read32: DWord; inline;
    function Read64: QWord; inline;
    function ReadAddress: TAddress; inline;
    function ReadAddressSigned: TAddressSigned; inline;
    constructor Create;
    destructor Destroy; override;
    property DataSizeBase: TBitWidth read FDataSizeBase write SetDataSizeBase;
    property AddrSizeBase: TBitWidth read FAddrSizeBase write SetAddrSizeBase;
    property Ticks: Integer read FTicks;
    property Running: Boolean read FRunning;
    property OnInput: TInputEvent read FOnInput write FOnInput;
    property OnOutput: TOutputEvent read FOnOutput write FOnOutput;
  end;

 { TCpuThread }

  TCpuThread = class(TThread)
    Cpu: TCpu;
    procedure Execute; override;
  end;

  TOpcodeParam = (prNone, prReg, prData, prAddr, prAddrRel);
  TOpcodeDef = record
    Name: string;
    Param1: TOpcodeParam;
    Param2: TOpcodeParam;
    Param3: TOpcodeParam;
    Prefix: Boolean;
  end;

const
  BitWidthBytes: array[TBitWidth] of Byte = (0, 1, 2, 4, 8);
  BitWidthText: array[TBitWidth] of string = ('None', '8-bit', '16-bit', '32-bit', '64-bit');
  OpcodeDef: array[TOpcode] of TOpcodeDef = (
    (Name: 'NOP'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'HALT'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'LD'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'LDI'; Param1: prReg; Param2: prData; Param3: prNone; Prefix: False),
    (Name: 'JP'; Param1: prAddr; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'JPZ'; Param1: prAddr; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'JPNZ'; Param1: prAddr; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'JR'; Param1: prAddrRel; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'JRZ'; Param1: prAddrRel; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'JRNZ'; Param1: prAddrRel; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'NEG'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'CLR'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'LDM'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'STM'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'EX'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'PUSH'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'POP'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'CALL'; Param1: prAddr; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'RET'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'ADD'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'ADDI'; Param1: prReg; Param2: prData; Param3: prNone; Prefix: False),
    (Name: 'SUB'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'SUBI'; Param1: prReg; Param2: prData; Param3: prNone; Prefix: False),
    (Name: 'INC'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'DEC'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'IN'; Param1: prReg; Param2: prAddr; Param3: prNone; Prefix: False),
    (Name: 'OUT'; Param1: prAddr; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'SHL'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'SHR'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'DP8'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'DP16'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'DP32'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'DP64'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'DS'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'AS'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'TEST'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: False),
    (Name: 'AND'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'OR'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'XOR'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'LDDR'; Param1: prReg; Param2: prReg; Param3: prReg; Prefix: False),
    (Name: 'LDDR'; Param1: prReg; Param2: prReg; Param3: prReg; Prefix: False),
    (Name: 'MUL'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'DIV'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'MOD'; Param1: prReg; Param2: prReg; Param3: prNone; Prefix: False),
    (Name: 'AP8'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'AP16'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'AP32'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'AP64'; Param1: prNone; Param2: prNone; Param3: prNone; Prefix: True),
    (Name: 'CON'; Param1: prReg; Param2: prNone; Param3: prNone; Prefix: True)
  );

implementation

{ TCpuThread }

procedure TCpuThread.Execute;
begin
  Cpu.Run;
end;

{ TCpu }

procedure TCpu.InstNop;
begin
end;

procedure TCpu.InstHalt;
begin
  Terminated := True;
end;

procedure TCpu.InstLoad;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R2].B;
    bw16: Registers[R1].W := Registers[R2].W;
    bw32: Registers[R1].D := Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R2].Q;
  end;
end;

procedure TCpu.InstLoadImmediate;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Read8;
    bw16: Registers[R].W := Read16;
    bw32: Registers[R].D := Read32;
    bw64: Registers[R].Q := Read64;
  end;
end;

procedure TCpu.InstJump;
begin
  IP := ReadAddress;
end;

procedure TCpu.InstJumpZero;
var
  Addr: TAddress;
begin
  Addr := ReadAddress;
  if Z then IP := Addr;
end;

procedure TCpu.InstJumpNotZero;
var
  Addr: TAddress;
begin
  Addr := ReadAddress;
  if not Z then IP := Addr;
end;

procedure TCpu.InstJumpRel;
begin
  IP := IP + ReadAddressSigned;
end;

procedure TCpu.InstJumpRelZero;
var
  Addr: TAddressSigned;
begin
  Addr := ReadAddressSigned;
  if Z then IP := IP + Addr;
end;

procedure TCpu.InstJumpRelNotZero;
var
  Addr: TAddressSigned;
begin
  Addr := ReadAddressSigned;
  if not Z then IP := IP + Addr;
end;

procedure TCpu.InstTest;
var
  Reg: TRegIndex;
begin
  Reg := Read8;
  case DataSize of
    bw8: Z := Registers[Reg].B = 0;
    bw16: Z := Registers[Reg].W = 0;
    bw32: Z := Registers[Reg].D = 0;
    bw64: Z := Registers[Reg].Q = 0;
  end;
end;

procedure TCpu.InstNeg;
var
  Reg: TRegIndex;
begin
  Reg := Read8;
  case DataSize of
    bw8: Registers[Reg].B := -Registers[Reg].B;
    bw16: Registers[Reg].W := -Registers[Reg].W;
    bw32: Registers[Reg].D := -Registers[Reg].D;
    bw64: Registers[Reg].Q := -Registers[Reg].Q;
  end;
end;

procedure TCpu.InstClear;
begin
  case DataSize of
    bw8: Registers[Read8].B := 0;
    bw16: Registers[Read8].W := 0;
    bw32: Registers[Read8].D := 0;
    bw64: Registers[Read8].Q := 0;
  end;
end;

procedure TCpu.InstLoadMem;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case AddrSize of
    bw8: case DataSize of
      bw8: Registers[R1].B := PByte(Memory + PByte(@Registers[R2])^)^;
      bw16: Registers[R1].W := PWord(Memory + PByte(@Registers[R2])^)^;
      bw32: Registers[R1].D := PDWord(Memory + PByte(@Registers[R2])^)^;
      bw64: Registers[R1].Q := PQWord(Memory + PByte(@Registers[R2])^)^;
    end;
    bw16: case DataSize of
      bw8: Registers[R1].B := PByte(Memory + PWord(@Registers[R2])^)^;
      bw16: Registers[R1].W := PWord(Memory + PWord(@Registers[R2])^)^;
      bw32: Registers[R1].D := PDWord(Memory + PWord(@Registers[R2])^)^;
      bw64: Registers[R1].Q := PQWord(Memory + PWord(@Registers[R2])^)^;
    end;
    bw32: case DataSize of
      bw8: Registers[R1].B := PByte(Memory + PDWord(@Registers[R2])^)^;
      bw16: Registers[R1].W := PWord(Memory + PDWord(@Registers[R2])^)^;
      bw32: Registers[R1].D := PDWord(Memory + PDWord(@Registers[R2])^)^;
      bw64: Registers[R1].Q := PQWord(Memory + PDWord(@Registers[R2])^)^;
    end;
    bw64: case DataSize of
      bw8: Registers[R1].B := PByte(Memory + PQWord(@Registers[R2])^)^;
      bw16: Registers[R1].W := PWord(Memory + PQWord(@Registers[R2])^)^;
      bw32: Registers[R1].D := PDWord(Memory + PQWord(@Registers[R2])^)^;
      bw64: Registers[R1].Q := PQWord(Memory + PQWord(@Registers[R2])^)^;
    end;
  end;
end;

procedure TCpu.InstStoreMem;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case AddrSize of
    bw8: case DataSize of
      bw8: PByte(Memory + PByte(@Registers[R1])^)^ := Registers[R2].B;
      bw16: PWord(Memory + PByte(@Registers[R1])^)^ := Registers[R2].W;
      bw32: PDWord(Memory + PByte(@Registers[R1])^)^ := Registers[R2].D;
      bw64: PQWord(Memory + PByte(@Registers[R1])^)^ := Registers[R2].Q;
    end;
    bw16: case DataSize of
      bw8: PByte(Memory + PWord(@Registers[R1])^)^ := Registers[R2].B;
      bw16: PWord(Memory + PWord(@Registers[R1])^)^ := Registers[R2].W;
      bw32: PDWord(Memory + PWord(@Registers[R1])^)^ := Registers[R2].D;
      bw64: PQWord(Memory + PWord(@Registers[R1])^)^ := Registers[R2].Q;
    end;
    bw32: case DataSize of
      bw8: PByte(Memory + PDWord(@Registers[R1])^)^ := Registers[R2].B;
      bw16: PWord(Memory + PDWord(@Registers[R1])^)^ := Registers[R2].W;
      bw32: PDWord(Memory + PDWord(@Registers[R1])^)^ := Registers[R2].D;
      bw64: PQWord(Memory + PDWord(@Registers[R1])^)^ := Registers[R2].Q;
    end;
    bw64: case DataSize of
      bw8: PByte(Memory + PQWord(@Registers[R1])^)^ := Registers[R2].B;
      bw16: PWord(Memory + PQWord(@Registers[R1])^)^ := Registers[R2].W;
      bw32: PDWord(Memory + PQWord(@Registers[R1])^)^ := Registers[R2].D;
      bw64: PQWord(Memory + PQWord(@Registers[R1])^)^ := Registers[R2].Q;
    end;
  end;
end;

procedure TCpu.InstExchg;
var
  Temp: TRegister;
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: begin
      Temp.B := Registers[R1].B;
      Registers[R1].B := Registers[R2].B;
      Registers[R2].B := Temp.B;
    end;
    bw16: begin
      Temp.W := Registers[R1].W;
      Registers[R1].W := Registers[R2].W;
      Registers[R2].W := Temp.W;
    end;
    bw32: begin
      Temp.D := Registers[R1].D;
      Registers[R1].D := Registers[R2].D;
      Registers[R2].D := Temp.D;
    end;
    bw64: begin
      Temp.q := Registers[R1].Q;
      Registers[R1].q := Registers[R2].Q;
      Registers[R2].Q := Temp.Q;
    end;
  end;
end;

procedure TCpu.InstPush;
begin
  case DataSize of
    bw8: begin
      Dec(SP, SizeOf(Byte));
      PByte(Memory + SP)^ := Registers[Read8].B;
    end;
    bw16: begin
      Dec(SP, SizeOf(Word));
      PWord(Memory + SP)^ := Registers[Read8].W;
    end;
    bw32: begin
      Dec(SP, SizeOf(DWord));
      PDWord(Memory + SP)^ := Registers[Read8].D;
    end;
    bw64: begin
      Dec(SP, SizeOf(QWord));
      PQWord(Memory + SP)^ := Registers[Read8].Q;
    end;
  end;
end;

procedure TCpu.InstPop;
begin
  case DataSize of
    bw8: begin
      Registers[Read8].B := PByte(Memory + SP)^;
      Inc(SP, SizeOf(Byte));
    end;
    bw16: begin
      Registers[Read8].W := PWord(Memory + SP)^;
      Inc(SP, SizeOf(Word));
    end;
    bw32: begin
      Registers[Read8].D := PDWord(Memory + SP)^;
      Inc(SP, SizeOf(DWord));
    end;
    bw64: begin
      Registers[Read8].Q := PQWord(Memory + SP)^;
      Inc(SP, SizeOf(QWord));
    end;
  end;
end;

procedure TCpu.InstCall;
var
  Dest: TRegister;
begin
  case AddrSize of
    bw8: begin
      Dest.B := Read8;
      Dec(SP, SizeOf(Byte));
      PByte(Memory + SP)^ := Byte(IP);
      IP := Dest.B;
    end;
    bw16: begin
      Dest.W := Read16;
      Dec(SP, SizeOf(Word));
      PWord(Memory + SP)^ := Word(IP);
      IP := Dest.W;
    end;
    bw32: begin
      Dest.D := Read32;
      Dec(SP, SizeOf(DWord));
      PDWord(Memory + SP)^ := DWord(IP);
      IP := Dest.D;
    end;
    bw64: begin
      Dest.Q := Read64;
      Dec(SP, SizeOf(QWord));
      PQWord(Memory + SP)^ := QWord(IP);
      IP := Dest.Q;
    end;
  end;
end;

procedure TCpu.InstRet;
begin
  case AddrSize of
    bw8: begin
      IP := PByte(Memory + SP)^;
      Inc(SP, SizeOf(Byte));
    end;
    bw16: begin
      IP := PWord(Memory + SP)^;
      Inc(SP, SizeOf(Word));
    end;
    bw32: begin
      IP := PDWord(Memory + SP)^;
      Inc(SP, SizeOf(DWord));
    end;
    bw64: begin
      IP := PQWord(Memory + SP)^;
      Inc(SP, SizeOf(QWord));
    end;
  end;
end;

procedure TCpu.InstAdd;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B + Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W + Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D + Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q + Registers[R2].Q;
  end;
end;

procedure TCpu.InstAddi;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Registers[R].B + Read8;
    bw16: Registers[R].W := Registers[R].W + Read16;
    bw32: Registers[R].D := Registers[R].D + Read32;
    bw64: Registers[R].Q := Registers[R].Q + Read64;
  end;
end;

procedure TCpu.InstSub;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B - Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W - Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D - Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q - Registers[R2].Q;
  end;
end;

procedure TCpu.InstSubi;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Registers[R].B - Read8;
    bw16: Registers[R].W := Registers[R].W - Read16;
    bw32: Registers[R].D := Registers[R].D - Read32;
    bw64: Registers[R].Q := Registers[R].Q - Read64;
  end;
end;

procedure TCpu.InstMul;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B * Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W * Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D * Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q * Registers[R2].Q;
  end;
end;

procedure TCpu.InstDiv;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B div Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W div Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D div Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q div Registers[R2].Q;
  end;
end;

procedure TCpu.InstMod;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B mod Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W mod Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D mod Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q mod Registers[R2].Q;
  end;
end;

procedure TCpu.InstInc;
var
  R1: TRegIndex;
begin
  R1 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B + 1;
    bw16: Registers[R1].W := Registers[R1].W + 1;
    bw32: Registers[R1].D := Registers[R1].D + 1;
    bw64: Registers[R1].Q := Registers[R1].Q + 1;
  end;
end;

procedure TCpu.InstDec;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Registers[R].B - 1;
    bw16: Registers[R].W := Registers[R].W - 1;
    bw32: Registers[R].D := Registers[R].D - 1;
    bw64: Registers[R].Q := Registers[R].Q - 1;
  end;
end;

procedure TCpu.InstIn;
var
  R: TRegIndex;
  Port: TAddress;
begin
  R := Read8;
  Port := ReadAddress;
  case DataSize of
    bw8: begin
      if Assigned(FOnInput) then Registers[R].B := FOnInput(Port).B
        else Registers[R].B := 0;
    end;
    bw16: begin
      if Assigned(FOnInput) then Registers[R].W := FOnInput(Port).W
        else Registers[R].W := 0;
    end;
    bw32: begin
      if Assigned(FOnInput) then Registers[R].D := FOnInput(Port).D
        else Registers[R].D := 0;
    end;
    bw64: begin
      if Assigned(FOnInput) then Registers[R].Q := FOnInput(Port).Q
        else Registers[R].Q := 0;
    end;
  end;
end;

procedure TCpu.InstOut;
var
  R: TRegIndex;
  Port: TAddress;
  Value: TRegister;
begin
  Port := ReadAddress;
  R := Read8;
  case DataSize of
    bw8: begin
      Value.B := Registers[R].B;
      if Assigned(FOnOutput) then
        FOnOutput(Port, Value);
    end;
    bw16: begin
      Value.W := Registers[R].W;
      if Assigned(FOnOutput) then
        FOnOutput(Port, Value);
    end;
    bw32: begin
      Value.D := Registers[R].D;
      if Assigned(FOnOutput) then
        FOnOutput(Port, Value);
    end;
    bw64: begin
      Value.Q := Registers[R].Q;
      if Assigned(FOnOutput) then
        FOnOutput(Port, Value);
    end;
  end;
end;

procedure TCpu.InstShr;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Registers[R].B shr Read8;
    bw16: Registers[R].W := Registers[R].W shr Read16;
    bw32: Registers[R].D := Registers[R].D shr Read32;
    bw64: Registers[R].Q := Registers[R].Q shr Read64;
  end;
end;

procedure TCpu.InstShl;
var
  R: TRegIndex;
begin
  R := Read8;
  case DataSize of
    bw8: Registers[R].B := Registers[R].B shl Read8;
    bw16: Registers[R].W := Registers[R].W shl Read16;
    bw32: Registers[R].D := Registers[R].D shl Read32;
    bw64: Registers[R].Q := Registers[R].Q shl Read64;
  end;
end;

procedure TCpu.InstAnd;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B and Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W and Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D and Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q and Registers[R2].Q;
  end;
end;

procedure TCpu.InstOr;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B or Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W or Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D or Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q or Registers[R2].Q;
  end;
end;

procedure TCpu.InstXor;
var
  R1, R2: TRegIndex;
begin
  R1 := Read8;
  R2 := Read8;
  case DataSize of
    bw8: Registers[R1].B := Registers[R1].B xor Registers[R2].B;
    bw16: Registers[R1].W := Registers[R1].W xor Registers[R2].W;
    bw32: Registers[R1].D := Registers[R1].D xor Registers[R2].D;
    bw64: Registers[R1].Q := Registers[R1].Q xor Registers[R2].Q;
  end;
end;

procedure TCpu.InstConvert;
var
  R: TRegIndex;
begin
  R := Read8;
  case AddrSize of
    bw8: case DataSize of
      bw8: Registers[R].B := Registers[R].B;
      bw16: Registers[R].B := Registers[R].W;
      bw32: Registers[R].B := Registers[R].D;
      bw64: Registers[R].B := Registers[R].Q;
    end;
    bw16: case DataSize of
      bw8: Registers[R].W := Registers[R].B;
      bw16: Registers[R].W := Registers[R].W;
      bw32: Registers[R].W := Registers[R].D;
      bw64: Registers[R].W := Registers[R].Q;
    end;
    bw32: case DataSize of
      bw8: Registers[R].D := Registers[R].B;
      bw16: Registers[R].D := Registers[R].W;
      bw32: Registers[R].D := Registers[R].D;
      bw64: Registers[R].D := Registers[R].Q;
    end;
    bw64: case DataSize of
      bw8: Registers[R].Q := Registers[R].B;
      bw16: Registers[R].Q := Registers[R].W;
      bw32: Registers[R].Q := Registers[R].D;
      bw64: Registers[R].Q := Registers[R].Q;
    end;
  end;
end;


procedure TCpu.InstLddr;
var
  Source, Dest, Count: TRegIndex;
begin
  Source := Read8;
  Dest := Read8;
  Count := Read8;
  case AddrSize of
    bw8: while Registers[Count].B > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].B)^ := PByte(Memory + Registers[Source].B)^;
          Dec(Registers[Dest].B, SizeOf(Byte));
          Dec(Registers[Source].B, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].B)^ := PWord(Memory + Registers[Source].B)^;
          Dec(Registers[Dest].B, SizeOf(Word));
          Dec(Registers[Source].B, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].B)^ := PDWord(Memory + Registers[Source].B)^;
          Dec(Registers[Dest].B, SizeOf(DWord));
          Dec(Registers[Source].B, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].B)^ := PQWord(Memory + Registers[Source].B)^;
          Dec(Registers[Dest].B, SizeOf(QWord));
          Dec(Registers[Source].B, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].B);
    end;
    bw16: while Registers[Count].W > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].W)^ := PByte(Memory + Registers[Source].W)^;
          Dec(Registers[Dest].W, SizeOf(Byte));
          Dec(Registers[Source].W, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].W)^ := PWord(Memory + Registers[Source].W)^;
          Dec(Registers[Dest].W, SizeOf(Word));
          Dec(Registers[Source].W, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].W)^ := PDWord(Memory + Registers[Source].W)^;
          Dec(Registers[Dest].W, SizeOf(DWord));
          Dec(Registers[Source].W, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].W)^ := PQWord(Memory + Registers[Source].W)^;
          Dec(Registers[Dest].W, SizeOf(QWord));
          Dec(Registers[Source].W, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].W);
    end;
    bw32: while Registers[Count].D > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].D)^ := PByte(Memory + Registers[Source].D)^;
          Dec(Registers[Dest].D, SizeOf(Byte));
          Dec(Registers[Source].D, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].D)^ := PWord(Memory + Registers[Source].D)^;
          Dec(Registers[Dest].D, SizeOf(Word));
          Dec(Registers[Source].D, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].W)^ := PDWord(Memory + Registers[Source].D)^;
          Dec(Registers[Dest].D, SizeOf(DWord));
          Dec(Registers[Source].D, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].D)^ := PQWord(Memory + Registers[Source].D)^;
          Dec(Registers[Dest].D, SizeOf(QWord));
          Dec(Registers[Source].D, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].D);
    end;
    bw64: while Registers[Count].Q > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].Q)^ := PByte(Memory + Registers[Source].Q)^;
          Dec(Registers[Dest].Q, SizeOf(Byte));
          Dec(Registers[Source].Q, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].Q)^ := PWord(Memory + Registers[Source].Q)^;
          Dec(Registers[Dest].Q, SizeOf(Word));
          Dec(Registers[Source].Q, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].Q)^ := PDWord(Memory + Registers[Source].Q)^;
          Dec(Registers[Dest].Q, SizeOf(DWord));
          Dec(Registers[Source].Q, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].Q)^ := PQWord(Memory + Registers[Source].Q)^;
          Dec(Registers[Dest].Q, SizeOf(QWord));
          Dec(Registers[Source].Q, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].W);
    end;
  end;
end;

procedure TCpu.InstLdir;
var
  Source, Dest, Count: TRegIndex;
begin
  Source := Read8;
  Dest := Read8;
  Count := Read8;
  case AddrSize of
    bw8: while Registers[Count].B > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].B)^ := PByte(Memory + Registers[Source].B)^;
          Inc(Registers[Dest].B, SizeOf(Byte));
          Inc(Registers[Source].B, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].B)^ := PWord(Memory + Registers[Source].B)^;
          Inc(Registers[Dest].B, SizeOf(Word));
          Inc(Registers[Source].B, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].B)^ := PDWord(Memory + Registers[Source].B)^;
          Inc(Registers[Dest].B, SizeOf(DWord));
          Inc(Registers[Source].B, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].B)^ := PQWord(Memory + Registers[Source].B)^;
          Inc(Registers[Dest].B, SizeOf(QWord));
          Inc(Registers[Source].B, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].B);
    end;
    bw16: while Registers[Count].W > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].W)^ := PByte(Memory + Registers[Source].W)^;
          Inc(Registers[Dest].W, SizeOf(Byte));
          Inc(Registers[Source].W, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].W)^ := PWord(Memory + Registers[Source].W)^;
          Inc(Registers[Dest].W, SizeOf(Word));
          Inc(Registers[Source].W, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].W)^ := PDWord(Memory + Registers[Source].W)^;
          Inc(Registers[Dest].W, SizeOf(DWord));
          Inc(Registers[Source].W, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].W)^ := PQWord(Memory + Registers[Source].W)^;
          Inc(Registers[Dest].W, SizeOf(QWord));
          Inc(Registers[Source].W, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].W);
    end;
    bw32: while Registers[Count].D > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].D)^ := PByte(Memory + Registers[Source].D)^;
          Inc(Registers[Dest].D, SizeOf(Byte));
          Inc(Registers[Source].D, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].D)^ := PWord(Memory + Registers[Source].D)^;
          Inc(Registers[Dest].D, SizeOf(Word));
          Inc(Registers[Source].D, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].W)^ := PDWord(Memory + Registers[Source].D)^;
          Inc(Registers[Dest].D, SizeOf(DWord));
          Inc(Registers[Source].D, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].D)^ := PQWord(Memory + Registers[Source].D)^;
          Inc(Registers[Dest].D, SizeOf(QWord));
          Inc(Registers[Source].D, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].D);
    end;
    bw64: while Registers[Count].Q > 0 do begin
      case DataSize of
        bw8: begin
          PByte(Memory + Registers[Dest].Q)^ := PByte(Memory + Registers[Source].Q)^;
          Inc(Registers[Dest].Q, SizeOf(Byte));
          Inc(Registers[Source].Q, SizeOf(Byte));
        end;
        bw16: begin
          PWord(Memory + Registers[Dest].Q)^ := PWord(Memory + Registers[Source].Q)^;
          Inc(Registers[Dest].Q, SizeOf(Word));
          Inc(Registers[Source].Q, SizeOf(Word));
        end;
        bw32: begin
          PDWord(Memory + Registers[Dest].Q)^ := PDWord(Memory + Registers[Source].Q)^;
          Inc(Registers[Dest].Q, SizeOf(DWord));
          Inc(Registers[Source].Q, SizeOf(DWord));
        end;
        bw64: begin
          PQWord(Memory + Registers[Dest].Q)^ := PQWord(Memory + Registers[Source].Q)^;
          Inc(Registers[Dest].Q, SizeOf(QWord));
          Inc(Registers[Source].Q, SizeOf(QWord));
        end;
      end;
      Dec(Registers[Count].W);
    end;
  end;
end;

procedure TCpu.InstDataPrefix8;
begin
  DataSize := bw8;
  Prefix := True;
end;

procedure TCpu.InstDataPrefix16;
begin
  DataSize := bw16;
  Prefix := True;
end;

procedure TCpu.InstDataPrefix32;
begin
  DataSize := bw32;
  Prefix := True;
end;

procedure TCpu.InstDataPrefix64;
begin
  DataSize := bw64;
  Prefix := True;
end;

procedure TCpu.InstDataSize;
begin
  DataSizeBase := TBitWidth(Read8);
  DataSize := DataSizeBase;
end;

procedure TCpu.InstAddrPrefix8;
begin
  AddrSize := bw8;
  Prefix := True;
end;

procedure TCpu.InstAddrPrefix16;
begin
  AddrSize := bw16;
  Prefix := True;
end;

procedure TCpu.InstAddrPrefix32;
begin
  AddrSize := bw32;
  Prefix := True;
end;

procedure TCpu.InstAddrPrefix64;
begin
  AddrSize := bw64;
  Prefix := True;
end;

procedure TCpu.InstAddrSize;
begin
  AddrSizeBase := TBitWidth(Read8);
  AddrSize := AddrSizeBase;
end;

procedure TCpu.InitInstructions;
begin
  Instructions[opNop] := InstNop;
  Instructions[opHalt] := InstHalt;
  Instructions[opLoad] := InstLoad;
  Instructions[opLoadi] := InstLoadImmediate;
  Instructions[opJump] := InstJump;
  Instructions[opJumpNotZero] := InstJumpNotZero;
  Instructions[opJumpZero] := InstJumpZero;
  Instructions[opJumpRel] := InstJumpRel;
  Instructions[opJumpRelNotZero] := InstJumpRelNotZero;
  Instructions[opJumpRelZero] := InstJumpRelZero;
  Instructions[opLoadMem] := InstLoadMem;
  Instructions[opStoreMem] := InstStoreMem;
  Instructions[opClear] := InstClear;
  Instructions[opNeg] := InstNeg;
  Instructions[opExchg] := InstExchg;
  Instructions[opPop] := InstPop;
  Instructions[opPush] := InstPush;
  Instructions[opCall] := InstCall;
  Instructions[opRet] := InstRet;
  Instructions[opAdd] := InstAdd;
  Instructions[opAddi] := InstAddi;
  Instructions[opSub] := InstSub;
  Instructions[opSubi] := InstSubi;
  Instructions[opInc] := InstInc;
  Instructions[opDec] := InstDec;
  Instructions[opIn] := InstIn;
  Instructions[opOut] := InstOut;
  Instructions[opShr] := InstShr;
  Instructions[opShl] := InstShl;
  Instructions[opDataSize] := InstDataSize;
  Instructions[opDataPrefix8] := InstDataPrefix8;
  Instructions[opDataPrefix16] := InstDataPrefix16;
  Instructions[opDataPrefix32] := InstDataPrefix32;
  Instructions[opDataPrefix64] := InstDataPrefix64;
  Instructions[opAddrSize] := InstAddrSize;
  Instructions[opAddrPrefix8] := InstAddrPrefix8;
  Instructions[opAddrPrefix16] := InstAddrPrefix16;
  Instructions[opAddrPrefix32] := InstAddrPrefix32;
  Instructions[opAddrPrefix64] := InstAddrPrefix64;
  Instructions[opTest] := InstTest;
  Instructions[opAnd] := InstAnd;
  Instructions[opOr] := InstOr;
  Instructions[opXor] := InstXor;
  Instructions[opLdir] := InstLdir;
  Instructions[opLddr] := InstLddr;
  Instructions[opMul] := InstMul;
  Instructions[opDiv] := InstDiv;
  Instructions[opMod] := InstMod;
  Instructions[opConvert] := InstConvert;
end;

procedure TCpu.SetAddrSizeBase(AValue: TBitWidth);
begin
  if FAddrSizeBase = AValue then Exit;
  FAddrSizeBase := AValue;
  AddrSize := AValue;
end;

procedure TCpu.SetDataSizeBase(AValue: TBitWidth);
begin
  if FDataSizeBase = AValue then Exit;
  FDataSizeBase := AValue;
  DataSize := AValue;
end;

procedure TCpu.Run;
begin
  DataSize := DataSizeBase;
  AddrSize := AddrSizeBase;
  Terminated := False;
  FTicks := 0;
  IP := 0;
  SP := MemSize(Memory);
  while not Terminated do
    Step;
end;

procedure TCpu.Step;
var
  Opcode: Byte;
begin
  Prefix := False;
  Opcode := Read8;
  if Opcode < Length(Instructions) then begin
    if Assigned(Instructions[TOpcode(Opcode)]) then Instructions[TOpcode(Opcode)]
      else raise Exception.Create('Missing instruction handler for opcode '+ IntToStr(Opcode));
  end else raise Exception.Create('Unsupported opcode ' + IntToStr(Opcode) + ' at address ' + IntToHex(IP - 1, 8) + '.');
  if not Prefix then begin
    DataSize := DataSizeBase;
    AddrSize := AddrSizeBase;
  end;
  IP := IP mod MemSize(Memory);
  Inc(FTicks);
end;

procedure TCpu.Start;
begin
  if not Running then begin
    Terminated := False;
    Thread := TCpuThread.Create(True);
    Thread.Cpu := Self;
    Thread.Start;
    FRunning := True;
  end;
end;

procedure TCpu.Stop;
begin
  if Running then begin
    Terminated := True;
    Thread.Terminate;
    Thread.WaitFor;
    FreeAndNil(Thread);
    FRunning := False;
  end;
end;

function TCpu.Read8: Byte;
begin
  Result := PByte(Memory + IP)^;
  Inc(IP);
end;

function TCpu.Read16: Word;
begin
  Result := PWord(Memory + IP)^;
  Inc(IP, SizeOf(Word));
end;

function TCpu.Read32: DWord;
begin
  Result := PDWord(Memory + IP)^;
  Inc(IP, SizeOf(DWord));
end;

function TCpu.Read64: QWord;
begin
  Result := PQWord(Memory + IP)^;
  Inc(IP, SizeOf(QWord));
end;

function TCpu.ReadAddress: TAddress;
begin
  case AddrSize of
    bw8: Result := Read8;
    bw16: Result := Read16;
    bw32: Result := Read32;
    bw64: Result := Read64;
  end;
end;

function TCpu.ReadAddressSigned: TAddressSigned;
begin
  case AddrSize of
    bw8: Result := ShortInt(Read8);
    bw16: Result := SmallInt(Read16);
    bw32: Result := Integer(Read32);
    bw64: Result := Int64(Read64);
  end;
end;

constructor TCpu.Create;
begin
  DataSizeBase := bw16;
  AddrSizeBase := bw16;
  SetLength(Registers, 32);
  InitInstructions;
end;

destructor TCpu.Destroy;
begin
  Stop;
  inherited Destroy;
end;

end.

