unit UDisassembler;

{$mode delphi}

interface

uses
  Classes, SysUtils, UMemory, fgl, UCpu, UInstructionReader, Math;

type
  TDisassemblerLine = class
    Address: Integer;
    Opcode: string;
    Instruction: string;
  end;

  { TDisassembler }

  TDisassembler = class(TInstructionReader)
  private
  public
    Output: TFPGObjectList<TDisassemblerLine>;
    procedure Process;
    constructor Create;
    destructor Destroy; override;
  end;


implementation

const
  SignText: array[TValueSign] of string = ('-', '', '+');

function SignedIntToHex(Value: Int64; Digits: Byte): string;
begin
  Result := SignText[Sign(Value)] + IntToHex(Abs(Value), Digits);
end;

{ TDisassembler }

procedure TDisassembler.Process;
var
  Opcode: Byte;
  Line: TDisassemblerLine;
  MemorySize: Integer;
  Data: QWord;
  Address: TAddress;
  AddressRel: TAddressSigned;
  Reg: TRegIndex;
begin
  Init;
  Output.Clear;

  MemorySize := MemSize(Cpu.Memory);
  while IP < MemorySize do begin
    Prefix := False;
    Opcode := Read8;
    if Opcode <= Integer(High(TOpcode)) then begin
      Prefix := OpcodeDef[TOpcode(Opcode)].Prefix;
      case TOpcode(Opcode) of
        opDataPrefix8: DataSize := bw8;
        opDataPrefix16: DataSize := bw16;
        opDataPrefix32: DataSize := bw32;
        opDataPrefix64: DataSize := bw64;
        opAddrPrefix8: AddrSize := bw8;
        opAddrPrefix16: AddrSize := bw16;
        opAddrPrefix32: AddrSize := bw32;
        opAddrPrefix64: AddrSize := bw64;
      end;
      Line := TDisassemblerLine.Create;
      Line.Address := IP - 1;
      Line.Opcode := IntToHex(Opcode, 2);
      Line.Instruction := OpcodeDef[TOpcode(Opcode)].Name;
      case OpcodeDef[TOpcode(Opcode)].Param1 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ' R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ' $' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ' $' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(QWord(AddressRel), BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ' $' + SignedIntToHex(AddressRel, BitWidthBytes[AddrSize] * 2);
        end;
      end;
      case OpcodeDef[TOpcode(Opcode)].Param2 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ', R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(AddressRel, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + SignedIntToHex(AddressRel, BitWidthBytes[AddrSize] * 2);
        end;
      end;
      case OpcodeDef[TOpcode(Opcode)].Param3 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ', R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(QWord(AddressRel), BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', $' + SignedIntToHex(AddressRel, BitWidthBytes[AddrSize] * 2);
        end;
      end;
      Output.Add(Line);
    end else begin
      {Line := TDisassemblerLine.Create;
      Line.Address := IP;
      Line.Opcode := IntToHex(Opcode, 2);
      Line.Instruction := '';
      Output.Add(Line);
      }
    end;
    if not Prefix then begin
      DataSize := DataSizeBase;
      AddrSize := AddrSizeBase;
    end;
  end;
end;

constructor TDisassembler.Create;
begin
  Output := TFPGObjectList<TDisassemblerLine>.Create;
end;

destructor TDisassembler.Destroy;
begin
  Output.Free;
  inherited Destroy;
end;


end.

